@@ -9,7 +9,7 @@
  * See the file "COPYING" for information on usage and redistribution
  * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
  *
- * @(#) $Id: check.c,v 1.36 2002/03/05 14:27:07 strauss Exp $
+ * @(#) $Id: check.c,v 1.52 2003/11/18 12:56:04 schoenw Exp $
  */
 
 #include <config.h>
@@ -357,7 +357,7 @@ smiCheckTypeName(Parser *parser, Module *module, char *name, int line)
  *      Check whether a format specification is valid.
  *
  * Results:
- *      None.
+ *      Returns 1 of the format is acceptable and 0 otherwise.
  *
  * Side effects:
  *      None.
@@ -365,7 +365,7 @@ smiCheckTypeName(Parser *parser, Module *module, char *name, int line)
  *----------------------------------------------------------------------
  */
 
-void
+int
 smiCheckFormat(Parser *parser, SmiBasetype basetype, char *format, int line)
 {
     int n, repeat, error = 1;
@@ -427,6 +427,8 @@ smiCheckFormat(Parser *parser, SmiBasetype basetype, char *format, int line)
 	    smiPrintError(parser, ERR_INVALID_FORMAT, format);
 	}
     }
+
+    return ! error;
 }
 
 
@@ -528,7 +530,7 @@ smiCheckNamedNumberSubtyping(Parser *parser, Type *type)
 	    && type->export.basetype != SMI_BASETYPE_BITS)) {
 	return;
     }
-    
+
     for (list1Ptr = type->listPtr;
 	 list1Ptr; list1Ptr = list1Ptr->nextPtr) {
 	
@@ -558,17 +560,30 @@ smiCheckNamedNumberSubtyping(Parser *parser, Type *type)
 	    
 	if (! list2Ptr) {
 	    if (type->export.basetype == SMI_BASETYPE_ENUM) {
-		smiPrintErrorAtLine(parser, ERR_ENUM_SUBTYPE,
-				    type->line,
-				    nn1Ptr->export.name,
-				    nn1Ptr->export.value.value.integer32,
-				    type->parentPtr->export.name);
+		if (type->parentPtr->export.name) {
+		    smiPrintErrorAtLine(parser, ERR_ENUM_SUBTYPE_OF,
+					type->line,
+					nn1Ptr->export.name,
+					nn1Ptr->export.value.value.integer32,
+					type->parentPtr->export.name);
+		} else {
+		    smiPrintErrorAtLine(parser, ERR_ENUM_SUBTYPE,
+					type->line,
+					nn1Ptr->export.name,
+					nn1Ptr->export.value.value.integer32);
+		}
 	    }
 	    if (type->export.basetype == SMI_BASETYPE_BITS) {
-		smiPrintErrorAtLine(parser, ERR_BITS_SUBTYPE,
-				    type->line,
-				    nn1Ptr->export.name,
-				    type->parentPtr->export.name);
+		if (type->parentPtr->export.name) {
+		    smiPrintErrorAtLine(parser, ERR_BITS_SUBTYPE_OF,
+					type->line,
+					nn1Ptr->export.name,
+					type->parentPtr->export.name);
+		} else {
+		    smiPrintErrorAtLine(parser, ERR_BITS_SUBTYPE,
+					type->line,
+					nn1Ptr->export.name);
+		}
 	    }
 	}
     }
@@ -579,11 +594,11 @@ smiCheckNamedNumberSubtyping(Parser *parser, Type *type)
 /*
  *----------------------------------------------------------------------
  *
- * smiCheckBits --
+ * smiCheckNamedNumbersOrder --
  *
  *      Check and normalize the order of named numbers in a bits
  *	or enumeration type.
-  *
+ *
  * Results:
  *      None.
  *
@@ -606,6 +621,18 @@ smiCheckNamedNumbersOrder(Parser *parser, Type *type)
 	return;
     }
 
+    /* Check whether the first bit has been given a name. */
+
+    if (type->export.basetype == SMI_BASETYPE_BITS) {
+	for (listPtr = type->listPtr; listPtr; listPtr = listPtr->nextPtr) {
+	    nnPtr = (NamedNumber *)(listPtr->ptr);
+	    if (nnPtr->export.value.value.unsigned32 == 0) break;
+	}
+	if (! listPtr) {
+	    smiPrintErrorAtLine(parser, ERR_BITS_ZERO_NOT_NAMED, type->line);
+	}
+    }
+
     lastPtr = NULL;
     for (listPtr = type->listPtr; listPtr; listPtr = nextPtr) {
 
@@ -706,7 +733,14 @@ smiCheckIndex(Parser *parser, Object *object)
     Type *typePtr, *rTypePtr;
     Range *rangePtr;
     NamedNumber *nnPtr;
+    Node *nodePtr;
     int minSize, maxSize, len = 0;
+    int aux = 0, cols = 0, acc = 0;
+
+    for (nodePtr = object->nodePtr->firstChildPtr, cols = 0;
+	 nodePtr; nodePtr = nodePtr->nextPtr) {
+	cols++;
+    }
 
     for (listPtr = object->listPtr; listPtr; listPtr = listPtr->nextPtr) {
 	
@@ -868,20 +902,11 @@ smiCheckIndex(Parser *parser, Object *object)
 				object->export.name);
 	}
 
-	/*
-	 * TODO: The test below should not be performed if the table
-	 * only contains INDEX objects. In this case, one of the objects
-	 * has to be accessible (see 7.7 in RFC 2578).
-	 */
-
-	/* A good test case is the SNMP-VIEW-BASED-ACM-MIB. */
-
-	if ((parser->modulePtr->export.language == SMI_LANGUAGE_SMIV2)
-	    && (indexPtr->nodePtr->parentPtr == object->nodePtr)) {
-	    if (indexPtr->export.access != SMI_ACCESS_NOT_ACCESSIBLE) {
-		smiPrintErrorAtLine(parser, ERR_INDEX_ACCESSIBLE,
-				    object->line,
-				    indexPtr->export.name, object->export.name);
+    	for (nodePtr = object->nodePtr->firstChildPtr;
+	     nodePtr; nodePtr = nodePtr->nextPtr) {
+	    if (indexPtr == nodePtr->lastObjectPtr) {
+		aux++;
+		break;
 	    }
 	}
     }
@@ -891,6 +916,40 @@ smiCheckIndex(Parser *parser, Object *object)
 			    object->export.name,
 			    (object->export.oidlen + 1 + len) - 128);
     }
+
+    /* RFC 2578 section 7.7: Auxiliary objects must be not-accessible except
+       in some interesting corner cases. */
+
+    for (listPtr = object->listPtr; listPtr; listPtr = listPtr->nextPtr) {
+	
+	indexPtr = (Object *) listPtr->ptr;
+	typePtr = indexPtr->typePtr;
+
+	if (aux < cols) {
+	    if ((parser->modulePtr->export.language == SMI_LANGUAGE_SMIV2)
+		&& (indexPtr->nodePtr->parentPtr == object->nodePtr)) {
+		if (indexPtr->export.access != SMI_ACCESS_NOT_ACCESSIBLE) {
+		    smiPrintErrorAtLine(parser, ERR_INDEX_ACCESSIBLE,
+					object->line,
+					indexPtr->export.name, object->export.name);
+		}
+	    }
+	}
+	
+	for (nodePtr = object->nodePtr->firstChildPtr, acc = 0;
+	     nodePtr; nodePtr = nodePtr->nextPtr) {
+	    if (indexPtr == nodePtr->lastObjectPtr
+		&& indexPtr->export.access != SMI_ACCESS_NOT_ACCESSIBLE) {
+		acc++;
+	    }
+	}
+    }
+
+    if ((parser->modulePtr->export.language == SMI_LANGUAGE_SMIV2)
+	&& aux == cols && acc != 1) {
+	smiPrintErrorAtLine(parser, ERR_INDEX_NON_ACCESSIBLE,
+			    object->line, object->export.name);
+    }
 }
 
 
@@ -926,22 +985,28 @@ smiCheckAugment(Parser *parser, Object *object)
 	return;
     }
     
-    if (object->relatedPtr->export.indexkind != SMI_INDEX_INDEX) {
+    if (object->relatedPtr->export.indexkind == SMI_INDEX_INDEX)
+        return;
+    
+    if (object->export.indexkind == SMI_INDEX_AUGMENT)
+    {
 	smiPrintErrorAtLine(parser, ERR_AUGMENT_NESTED, object->line,
 			    object->export.name,
 			    object->relatedPtr->export.name);
 	return;
     }
     
+    if (object->relatedPtr->export.indexkind != SMI_INDEX_SPARSE) {
+	smiPrintErrorAtLine(parser, ERR_EXTENDS_WRONG_ROW_TYPE, object->line,
+			    object->export.name,
+			    object->relatedPtr->export.name);
+	return;
+    }
+    
     /*
      * TODO: Check the size of the instance identifier and the OID
      * for this entry node.
      */
-
-#if 0
-    fprintf(stderr, "** %s augments -> %s\n", object->export.name,
-	    object->relatedPtr->export.name);
-#endif
 }
 
 
@@ -1107,6 +1172,148 @@ smiCheckTypeRanges(Parser *parser, Type *type)
 
 
 
+/*
+ *----------------------------------------------------------------------
+ *
+ * smiCheckTypeFormat --
+ *
+ *      Check whether we know a format specification for integer types.
+ *
+ * Results:
+ *      None.
+ *
+ * Side effects:
+ *      None.
+ *
+ *----------------------------------------------------------------------
+ */
+
+void
+smiCheckTypeFormat(Parser *parser, Type *type)
+{
+    Type *t;
+    
+    if (! type || !type->export.name) {
+	return;
+    }
+
+    if (type->export.basetype != SMI_BASETYPE_INTEGER32
+	&& type->export.basetype != SMI_BASETYPE_INTEGER64
+	&& type->export.basetype != SMI_BASETYPE_UNSIGNED32
+	&& type->export.basetype != SMI_BASETYPE_UNSIGNED64) {
+	return;
+    }
+
+    for (t = type; t; t = t->parentPtr) {
+	if (t->export.format) {
+	    break;
+	}
+    }
+
+    if (! t) {
+	smiPrintErrorAtLine(parser, ERR_TYPE_WITHOUT_FORMAT, type->line,
+			    type->export.name);
+    }
+}
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * smiCheckValueType --
+ *
+ *      Check whether a given value matches a given type.
+ *
+ * Results:
+ *      None.
+ *
+ * Side effects:
+ *      None.
+ *
+ *----------------------------------------------------------------------
+ */
+
+void
+smiCheckValueType(Parser *parser, SmiValue *value, Type *type, int line)
+{
+    List *p, *nextPtr;
+    
+    if (value && (value->basetype != SMI_BASETYPE_UNKNOWN) && type) {
+
+	/*
+	 * If defval type and object type don't match, check whether
+	 * the defval value is in the allowed range of the object's basetype.
+	 */
+	if ((type->export.basetype == SMI_BASETYPE_INTEGER32) ||
+	    (type->export.basetype == SMI_BASETYPE_ENUM)) {
+	    if (((value->basetype == SMI_BASETYPE_INTEGER64) &&
+		 ((value->value.integer64 < (SmiInteger64)(-2147483647 - 1)) ||
+		  (value->value.integer64 > (SmiInteger64)2147483647))) ||
+		((value->basetype == SMI_BASETYPE_UNSIGNED32) &&
+		 ((value->value.unsigned32 > 2147483647))) ||
+		((value->basetype == SMI_BASETYPE_UNSIGNED64) &&
+		 ((value->value.unsigned32 > 2147483647)))) {
+		smiPrintErrorAtLine(parser, ERR_DEFVAL_OUT_OF_BASETYPE, line);
+	    }
+	}
+	if (type->export.basetype == SMI_BASETYPE_UNSIGNED32) {
+	    if (((value->basetype == SMI_BASETYPE_INTEGER64) &&
+		 ((value->value.integer64 < 0) ||
+		  (value->value.integer64 > (SmiInteger64)4294967295UL))) ||
+		((value->basetype == SMI_BASETYPE_INTEGER32) &&
+		 ((value->value.integer32 < 0))) ||
+		((value->basetype == SMI_BASETYPE_UNSIGNED64) &&
+		 ((value->value.unsigned32 > (SmiUnsigned32)4294967295UL)))) {
+		smiPrintErrorAtLine(parser, ERR_DEFVAL_OUT_OF_BASETYPE, line);
+	    }
+	}
+
+	/*
+	 * "cast" the defval to the object's basetype.
+	 */
+	value->basetype = type->export.basetype;
+
+	/*
+	 * check whether the defval matches the object's range restriction.
+	 */
+	if ((value->basetype == SMI_BASETYPE_UNSIGNED32) ||
+	    (value->basetype == SMI_BASETYPE_UNSIGNED64) ||
+	    (value->basetype == SMI_BASETYPE_INTEGER32) ||
+	    (value->basetype == SMI_BASETYPE_INTEGER64)) {
+	    for (p = type->listPtr; p; p = nextPtr) {
+		nextPtr = p->nextPtr;
+		if ((compareValues(&((Range *)p->ptr)->export.minValue,
+				   value) <= 0) &&
+		    (compareValues(&((Range *)p->ptr)->export.maxValue,
+				   value) >= 0)) {
+		    break;
+		}
+	    }
+	    if ((p == NULL) && type->listPtr) {
+		smiPrintErrorAtLine(parser, ERR_DEFVAL_OUT_OF_RANGE, line);
+	    }
+	}
+
+	/*
+	 * check whether the defval matches the object's enumeration.
+	 */
+	if (value->basetype == SMI_BASETYPE_ENUM) {
+	    for (p = type->listPtr; p; p = nextPtr) {
+		nextPtr = p->nextPtr;
+
+		if (((NamedNumber *)(p->ptr))->export.value.value.integer32 ==
+		    value->value.integer32) {
+			break;
+		}
+	    }
+	    if (p == NULL) {
+		smiPrintErrorAtLine(parser, ERR_DEFVAL_OUT_OF_ENUM, line);
+	    }
+	}
+    }
+}
+
+
+
 /*
  *----------------------------------------------------------------------
  *
@@ -1128,92 +1335,9 @@ void
 smiCheckDefault(Parser *parser, Object *object)
 {
     List *p, *nextPtr;
-    
-    if (object->export.value.basetype != SMI_BASETYPE_UNKNOWN
-	&& object->typePtr) {
 
-	/*
-	 * If defval type and object type don't match, check whether
-	 * the defval value is in the allowed range of the object's basetype.
-	 */
-	if ((object->typePtr->export.basetype == SMI_BASETYPE_INTEGER32) ||
-	    (object->typePtr->export.basetype == SMI_BASETYPE_ENUM)) {
-	    if (((object->export.value.basetype ==
-		  SMI_BASETYPE_INTEGER64) &&
-		 ((object->export.value.value.integer64 < (SmiInteger64)(-2147483647 - 1)) ||
-		  (object->export.value.value.integer64 > (SmiInteger64)2147483647))) ||
-		((object->export.value.basetype ==
-		  SMI_BASETYPE_UNSIGNED32) &&
-		 ((object->export.value.value.unsigned32 > 2147483647))) ||
-		((object->export.value.basetype ==
-		  SMI_BASETYPE_UNSIGNED64) &&
-		 ((object->export.value.value.unsigned32 > 2147483647)))) {
-		smiPrintErrorAtLine(parser, ERR_DEFVAL_OUT_OF_BASETYPE,
-				    object->line);
-	    }
-	}
-	if (object->typePtr->export.basetype == SMI_BASETYPE_UNSIGNED32) {
-	    if (((object->export.value.basetype ==
-		  SMI_BASETYPE_INTEGER64) &&
-		 ((object->export.value.value.integer64 < 0) ||
-		  (object->export.value.value.integer64 > (SmiInteger64)4294967295UL))) ||
-		((object->export.value.basetype ==
-		  SMI_BASETYPE_INTEGER32) &&
-		 ((object->export.value.value.integer32 < 0))) ||
-		((object->export.value.basetype ==
-		  SMI_BASETYPE_UNSIGNED64) &&
-		 ((object->export.value.value.unsigned32 > (SmiUnsigned32)4294967295UL)))) {
-		smiPrintErrorAtLine(parser, ERR_DEFVAL_OUT_OF_BASETYPE,
-				    object->line);
-	    }
-	}
-
-	/*
-	 * "cast" the defval to the object's basetype.
-	 */
-	object->export.value.basetype = object->typePtr->export.basetype;
-
-	/*
-	 * check whether the defval matches the object's range restriction.
-	 */
-	if ((object->export.value.basetype == SMI_BASETYPE_UNSIGNED32) ||
-	    (object->export.value.basetype == SMI_BASETYPE_UNSIGNED64) ||
-	    (object->export.value.basetype == SMI_BASETYPE_INTEGER32) ||
-	    (object->export.value.basetype == SMI_BASETYPE_INTEGER64)) {
-	    for (p = object->typePtr->listPtr; p; p = nextPtr) {
-		nextPtr = p->nextPtr;
-		if ((compareValues(&((Range *)p->ptr)->export.minValue,
-				   &object->export.value) <= 0) &&
-		    (compareValues(&((Range *)p->ptr)->export.maxValue,
-				   &object->export.value) >= 0)) {
-		    break;
-		}
-	    }
-	    if ((p == NULL) && object->typePtr->listPtr) {
-		smiPrintErrorAtLine(parser, ERR_DEFVAL_OUT_OF_RANGE,
-				    object->line);
-	    }
-	}
-
-	/*
-	 * check whether the defval matches the object's enumeration.
-	 */
-	if (object->export.value.basetype == SMI_BASETYPE_ENUM) {
-	    for (p = object->typePtr->listPtr; p; p = nextPtr) {
-		nextPtr = p->nextPtr;
-
-		if (((NamedNumber *)(p->ptr))->export.value.value.integer32 ==
-		    object->export.value.value.integer32) {
-			break;
-		}
-	    }
-	    if (p == NULL) {
-		smiPrintErrorAtLine(parser, ERR_DEFVAL_OUT_OF_ENUM,
-				    object->line);
-	    }
-	    
-	}
-    }
+    smiCheckValueType(parser, &object->export.value, object->typePtr,
+		      object->line);
 }
 
 
@@ -1234,21 +1358,215 @@ smiCheckDefault(Parser *parser, Object *object)
  *----------------------------------------------------------------------
  */
 
+static void
+checkInetAddressType(Parser *parserPtr, Module *modulePtr, Object *objectPtr)
+{
+    Module *inetModulePtr = NULL;		/* RFC 3291 */
+    Type *inetAddressTypePtr = NULL;		/* RFC 3291 */
+    Type *inetAddressPtr = NULL;		/* RFC 3291 */
+    Node *nodePtr;
+    List *listPtr;
+    int i;
+
+    const char *protected[] = {
+	"InetAddressIPv4", "InetAddressIPv6",
+	"InetAddressIPv4z", "InetAddressIPv6z",
+	"InetAddressDNS", NULL
+    };
+
+    inetModulePtr = findModuleByName("INET-ADDRESS-MIB");
+    if (! inetModulePtr) {
+	return;
+    }
+
+    inetAddressTypePtr = findTypeByModuleAndName(inetModulePtr,
+						 "InetAddressType");
+    inetAddressPtr = findTypeByModuleAndName(inetModulePtr,
+					     "InetAddress");
+    
+    if (!inetAddressTypePtr || !inetAddressPtr) {
+	return;
+    }
+    
+    /* check InetAddressType/InetAddress pair */
+    if (smiTypeDerivedFrom(objectPtr->typePtr, inetAddressPtr)) {
+	Object *indexObject = NULL;
+	Object *entryObject = objectPtr->nodePtr->parentPtr->lastObjectPtr;
+	if (entryObject) {
+	    switch (entryObject->export.indexkind) {
+	    case SMI_INDEX_INDEX:
+		indexObject = entryObject;
+		break;
+	    case SMI_INDEX_AUGMENT:
+		indexObject = entryObject->relatedPtr;
+		break;
+	    default:
+		/* xxx need to handle other index constructions */
+		indexObject = NULL;
+		break;
+	    }
+	}
+	if (indexObject) {
+	    for (listPtr = indexObject->listPtr; listPtr;
+		 listPtr = listPtr->nextPtr) {
+		Object *iObject = (Object *) listPtr->ptr;
+		if (iObject
+		    && iObject->typePtr == inetAddressTypePtr) {
+		    break;
+		}
+	    }
+	    
+	}
+	if (! indexObject || ! listPtr) {
+	    for (nodePtr =
+		     objectPtr->nodePtr->parentPtr->firstChildPtr;
+		 nodePtr &&
+		     nodePtr->subid < objectPtr->nodePtr->subid &&
+		     nodePtr->lastObjectPtr->typePtr !=
+		     inetAddressTypePtr;
+		 nodePtr = nodePtr->nextPtr);
+	    if (!nodePtr ||
+		nodePtr->subid >= objectPtr->nodePtr->subid) {
+		smiPrintErrorAtLine(parserPtr,
+				    ERR_INETADDRESS_WITHOUT_TYPE,
+				    objectPtr->line);
+	    }
+	}
+    }
+    
+    /* check InetAddressType subtyping */
+    if (objectPtr->typePtr->parentPtr == inetAddressTypePtr) {
+	smiPrintErrorAtLine(parserPtr,
+			    ERR_INETADDRESSTYPE_SUBTYPED,
+			    objectPtr->line);
+    }
+
+    /* check for TCs that should not be used directly */
+    for (i = 0; protected[i]; i++) {
+	if (objectPtr->typePtr == findTypeByModuleAndName(inetModulePtr,
+							  protected[i])) {
+	    smiPrintErrorAtLine(parserPtr,
+				ERR_INETADDRESS_SPECIFIC,
+				objectPtr->line,
+				objectPtr->typePtr->export.name);
+	    break;
+	}
+    }
+}
+
+
+static void
+checkTransportAddressType(Parser *parserPtr, Module *modulePtr, Object *objectPtr)
+{
+    Module *transportModulePtr = NULL;		/* RFC 3419 */
+    Type *transportAddressTypePtr = NULL;	/* RFC 3419 */
+    Type *transportAddressPtr = NULL;		/* RFC 3419 */
+    Type *transportDomainPtr = NULL;		/* RFC 3419 */
+    Node *nodePtr;
+    List *listPtr;
+    int i;
+
+    const char *protected[] = {
+	"TransportAddressIPv4", "TransportAddressIPv6",
+	"TransportAddressIPv4z", "TransportAddressIPv6z",
+	"TransportAddressDNS", "TransportAddressLocal",
+	NULL
+    };
+
+    transportModulePtr = findModuleByName("TRANSPORT-ADDRESS-MIB");
+    if (! transportModulePtr) {
+	return;
+    }
+
+    transportAddressTypePtr = findTypeByModuleAndName(transportModulePtr,
+						 "TransportAddressType");
+    transportAddressPtr = findTypeByModuleAndName(transportModulePtr,
+					     "TransportAddress");
+    transportDomainPtr = findTypeByModuleAndName(transportModulePtr,
+						 "TransportDomain");
+
+    if (!transportAddressTypePtr || !transportAddressPtr || !transportDomainPtr) {
+	return;
+    }
+    
+    /* check TransportAddressType/TransportAddress pair */
+    if (smiTypeDerivedFrom(objectPtr->typePtr, transportAddressPtr)) {
+	Object *indexObject = NULL;
+	Object *entryObject = objectPtr->nodePtr->parentPtr->lastObjectPtr;
+	if (entryObject) {
+	    switch (entryObject->export.indexkind) {
+	    case SMI_INDEX_INDEX:
+		indexObject = entryObject;
+		break;
+	    case SMI_INDEX_AUGMENT:
+		indexObject = entryObject->relatedPtr;
+		break;
+	    default:
+		/* xxx need to handle other index constructions */
+		indexObject = NULL;
+		break;
+	    }
+	}
+	if (indexObject) {
+	    for (listPtr = indexObject->listPtr; listPtr;
+		 listPtr = listPtr->nextPtr) {
+		Object *iObject = (Object *) listPtr->ptr;
+		if (iObject
+		    && (iObject->typePtr == transportAddressTypePtr
+			|| iObject->typePtr == transportDomainPtr)) {
+		    break;
+		}
+	    }
+	}
+	if (! indexObject || ! listPtr) {
+	    for (nodePtr =
+		     objectPtr->nodePtr->parentPtr->firstChildPtr;
+		 nodePtr &&
+		     nodePtr->subid < objectPtr->nodePtr->subid &&
+		     nodePtr->lastObjectPtr->typePtr !=
+		     transportAddressTypePtr &&
+		     nodePtr->lastObjectPtr->typePtr !=
+		     transportDomainPtr;
+		 nodePtr = nodePtr->nextPtr);
+	    if (!nodePtr ||
+		nodePtr->subid >= objectPtr->nodePtr->subid) {
+		smiPrintErrorAtLine(parserPtr,
+				    ERR_TRANSPORTADDRESS_WITHOUT_TYPE,
+				    objectPtr->line);
+	    }
+	}
+    }
+    
+    /* check TransportAddressType subtyping */
+    if (objectPtr->typePtr->parentPtr == transportAddressTypePtr) {
+	smiPrintErrorAtLine(parserPtr,
+			    ERR_TRANSPORTADDRESSTYPE_SUBTYPED,
+			    objectPtr->line);
+    }
+
+    /* check for TCs that should not be used directly */
+    for (i = 0; protected[i]; i++) {
+	if (objectPtr->typePtr == findTypeByModuleAndName(transportModulePtr,
+							  protected[i])) {
+	    smiPrintErrorAtLine(parserPtr,
+				ERR_TRANSPORTADDRESS_SPECIFIC,
+				objectPtr->line,
+				objectPtr->typePtr->export.name);
+	    break;
+	}
+    }
+}
+
+
 void
 smiCheckTypeUsage(Parser *parserPtr, Module *modulePtr)
 {
     Object *objectPtr;
     Module *tcModulePtr = NULL;
-    Module *inetModulePtr = NULL;
     Type *rowStatusPtr = NULL;
     Type *storageTypePtr = NULL;
     Type *taddressPtr = NULL;
     Type *tdomainPtr = NULL;
-    Type *inetAddressTypePtr = NULL;
-    Type *inetAddressPtr = NULL;
-    Type *inetAddressIPv4Ptr = NULL;
-    Type *inetAddressIPv6Ptr = NULL;
-    Type *inetAddressDNSPtr = NULL;
     NamedNumber *nnPtr;
     Node *nodePtr;
     
@@ -1259,26 +1577,11 @@ smiCheckTypeUsage(Parser *parserPtr, Module *modulePtr)
 	taddressPtr = findTypeByModuleAndName(tcModulePtr, "TAddress");
 	tdomainPtr = findTypeByModuleAndName(tcModulePtr, "TDomain");
     }
-    inetModulePtr = findModuleByName("INET-ADDRESS-MIB");
-    if (inetModulePtr) {
-	inetAddressTypePtr = findTypeByModuleAndName(inetModulePtr,
-						     "InetAddressType");
-	inetAddressPtr = findTypeByModuleAndName(inetModulePtr,
-						 "InetAddress");
-	inetAddressIPv4Ptr = findTypeByModuleAndName(inetModulePtr,
-						     "InetAddressIPv4");
-	inetAddressIPv6Ptr = findTypeByModuleAndName(inetModulePtr,
-						     "InetAddressIPv6");
-	inetAddressDNSPtr = findTypeByModuleAndName(inetModulePtr,
-						    "InetAddressDNS");
-    }
-    
-    if (!tcModulePtr && !inetModulePtr) return;
     
     for (objectPtr = modulePtr->firstObjectPtr;
 	 objectPtr; objectPtr = objectPtr->nextPtr) {
 
-	if (objectPtr->typePtr && objectPtr->typePtr) {
+	if (objectPtr->typePtr) {
 
 	    if (tcModulePtr) {
 
@@ -1324,50 +1627,13 @@ smiCheckTypeUsage(Parser *parserPtr, Module *modulePtr)
 		}
 		
 	    }
-	    if (inetModulePtr) {
-
-		/* check InetAddressType/InetAddress pair */
-		if (smiTypeDerivedFrom(objectPtr->typePtr, inetAddressPtr)) {
-		    for (nodePtr =
-			     objectPtr->nodePtr->parentPtr->firstChildPtr;
-			 nodePtr &&
-			     nodePtr->subid < objectPtr->nodePtr->subid &&
-			     nodePtr->lastObjectPtr->typePtr !=
-			         inetAddressTypePtr;
-			 nodePtr = nodePtr->nextPtr);
-		    if (!nodePtr ||
-			nodePtr->subid >= objectPtr->nodePtr->subid) {
-			smiPrintErrorAtLine(parserPtr,
-					    ERR_INETADDRESS_WITHOUT_TYPE,
-					    objectPtr->line);
-		    }
-		}
-
-		/* check InetAddressType subtyping */
-		if (objectPtr->typePtr->parentPtr == inetAddressTypePtr) {
-		    smiPrintErrorAtLine(parserPtr,
-					ERR_INETADDRESSTYPE_SUBTYPED,
-					objectPtr->line);
-		}
-
-		/* check InetAddressXXX usage  */
-		if (objectPtr->typePtr == inetAddressIPv4Ptr ||
-		    objectPtr->typePtr == inetAddressIPv6Ptr ||
-		    objectPtr->typePtr == inetAddressDNSPtr) {
-		    smiPrintErrorAtLine(parserPtr,
-					ERR_INETADDRESS_SPECIFIC,
-					objectPtr->line,
-					objectPtr->typePtr->export.name);
-		}
-
-	    }
-	    
+	    checkInetAddressType(parserPtr, modulePtr, objectPtr);
+	    checkTransportAddressType(parserPtr, modulePtr, objectPtr);
 	}
     }
 }
 
 
-
 static char *status[] = { "Unknown", "current", "deprecated",
 			  "mandatory", "optional", "obsolete" };
 
@@ -1433,7 +1699,7 @@ smiCheckComplianceStatus(Parser *parser, Object *compliance)
 	addObjectFlags(memberPtr, FLAG_INCOMPLIANCE);
 	if (memberPtr->export.status > compliance->export.status) {
 	    smiPrintErrorAtLine(parser, ERR_COMPLIANCE_GROUP_STATUS,
-				compliance->line,
+				((Option *) listPtr->ptr)->line,
 				status[compliance->export.status],
 				compliance->export.name,
 				status[memberPtr->export.status],
@@ -1449,7 +1715,7 @@ smiCheckComplianceStatus(Parser *parser, Object *compliance)
 	     groupListPtr; groupListPtr = groupListPtr->nextPtr) {
 
 	    group = (Object *) groupListPtr->ptr;
-	    if (memberOfGroup(memberPtr, group)) {
+	    if (group && memberOfGroup(memberPtr, group)) {
 		break;
 	    }
 	}
@@ -1459,7 +1725,7 @@ smiCheckComplianceStatus(Parser *parser, Object *compliance)
 		 groupListPtr; groupListPtr = groupListPtr->nextPtr) {
 
 		group = ((Option *) groupListPtr->ptr)->objectPtr;
-		if (memberOfGroup(memberPtr, group)) {
+		if (group && memberOfGroup(memberPtr, group)) {
 		    break;
 		}
 	    }
@@ -1467,14 +1733,14 @@ smiCheckComplianceStatus(Parser *parser, Object *compliance)
 
 	if (! groupListPtr) {
 	    smiPrintErrorAtLine(parser, ERR_REFINEMENT_NOT_LISTED,
-				compliance->line,
+				((Refinement *) listPtr->ptr)->line,
 				memberPtr->export.name);
 	}
 		
 	addObjectFlags(memberPtr, FLAG_INCOMPLIANCE);
 	if (memberPtr->export.status > compliance->export.status) {
 	    smiPrintErrorAtLine(parser, ERR_COMPLIANCE_OBJECT_STATUS,
-				compliance->line,
+				((Refinement *) listPtr->ptr)->line,
 				status[compliance->export.status],
 				compliance->export.name,
 				status[memberPtr->export.status],
@@ -1600,7 +1866,7 @@ smiCheckGroupMembership(Parser *parser, Object *objectPtr)
 	found = (objectPtr->flags & FLAG_INCOMPLIANCE);
 
 	if (!found && objectPtr->export.status != SMI_STATUS_OBSOLETE) {
-	    smiPrintErrorAtLine(parser, ERR_GROUP_OPTIONAL,
+	    smiPrintErrorAtLine(parser, ERR_GROUP_UNREF,
 				objectPtr->line,
 				status[objectPtr->export.status],
 				objectPtr->export.name);
@@ -1771,6 +2037,59 @@ smiCheckNotificationMembers(Parser *parser, Object *object)
     }
 }
 
+/*
+ *----------------------------------------------------------------------
+ *
+ * smiCheckUniqueness --
+ *
+ *      Check whether all entries for an UNIQUENESS clause are in fact
+ *      columns.
+ *
+ * Results:
+ *      None.
+ *
+ * Side effects:
+ *      None.
+ *
+ *----------------------------------------------------------------------
+ */
+
+void smiCheckUniqueness(Parser *parser, Object *object)
+{
+    List *p;
+
+    for (p = object->uniquenessPtr; p; p = p->nextPtr) {
+        Object *uniq = (Object *)p->ptr;
+        int found = 0;
+        List *pp;
+
+        if (uniq && object->typePtr) {
+            for (pp = object->typePtr->listPtr; pp; pp = pp->nextPtr)
+                if (pp->ptr &&
+                    !strcmp(uniq->export.name, ((Object *)pp->ptr)->export.name)) {
+                    found = 1;
+                    break;
+                }
+            if (!found) {
+                if (((object->export.indexkind == SMI_INDEX_AUGMENT) ||
+                    (object->export.indexkind == SMI_INDEX_SPARSE)) &&
+                    (object->relatedPtr && object->relatedPtr->typePtr)) {
+                    for (pp = object->relatedPtr->typePtr->listPtr; pp;
+                         pp = pp->nextPtr)
+                        if (pp->ptr &&
+                            !strcmp(uniq->export.name, ((Object *)pp->ptr)->export.name)) {
+                            found = 1;
+                            break;
+                        }
+                }
+            }
+            if (!found)
+                smiPrintErrorAtLine(parser, ERR_NOT_A_COLUMN,
+                                    object->line, uniq->export.name);
+        }
+    }
+}
+
 /*
  *----------------------------------------------------------------------
  *
