@@ -3,12 +3,12 @@
  *
  *      Operations on the main data structures.
  *
- * Copyright (c) 1999 Frank Strauss, Technical University of Braunschweig.
+ * Copyright (c) 1999-2002 Frank Strauss, Technical University of Braunschweig.
  *
  * See the file "COPYING" for information on usage and redistribution
  * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
  *
- * @(#) $Id: data.c,v 1.109 2001/12/17 18:05:22 schoenw Exp $
+ * @(#) $Id: data.c,v 1.120 2003/10/02 11:57:41 strauss Exp $
  */
 
 #include <config.h>
@@ -17,7 +17,7 @@
 #include <errno.h>
 #include <ctype.h>
 #include <sys/types.h>
-#if !defined(_MSC_VER)
+#if !defined(_MSC_VER) && !defined(__MINGW32__)
 #include <sys/wait.h>
 #endif
 #ifdef HAVE_UNISTD_H
@@ -225,9 +225,11 @@ int isInView(const char *modulename)
 {
     View	      *viewPtr;
 
+#if 0
     if (smiHandle->flags & SMI_FLAG_VIEWALL) {
 	return 1;
     }
+#endif
     
     for (viewPtr = smiHandle->firstViewPtr; viewPtr; viewPtr = viewPtr->nextPtr) {
 	if (!strcmp(modulename, viewPtr->name)) {
@@ -598,6 +600,7 @@ Import *addImport(char *name, Parser *parserPtr)
     importPtr->export.module		 = NULL; /* not yet known */
     importPtr->kind			 = KIND_UNKNOWN; /* not yet known */
     importPtr->use			 = 0;
+    importPtr->flags			 = 0;
     importPtr->line			 = parserPtr ? parserPtr->line : -1;
     
     importPtr->nextPtr			 = NULL;
@@ -613,6 +616,29 @@ Import *addImport(char *name, Parser *parserPtr)
 
 
 
+/*
+ *----------------------------------------------------------------------
+ *
+ * addImportFlags --
+ *
+ *      Add flags to the flags of a given Import struct.
+ *
+ * Results:
+ *	None.
+ *
+ * Side effects:
+ *      None.
+ *
+ *----------------------------------------------------------------------
+ */
+
+void addImportFlags(Import *importPtr, ImportFlags flags)
+{
+    importPtr->flags |= flags;
+}
+
+
+
 /*
  *----------------------------------------------------------------------
  *
@@ -733,7 +759,8 @@ Import *findImportByName(const char *name, Module *modulePtr)
     
     for (importPtr = modulePtr->firstImportPtr; importPtr;
 	 importPtr = importPtr->nextPtr) {
-	if (!strcmp(importPtr->export.name, name)) {
+	if ((!strcmp(importPtr->export.name, name)) &&
+	    (!(importPtr->flags & FLAG_INCOMPLIANCE))) {
 		return (importPtr);
 	}
     }
@@ -833,6 +860,8 @@ Object *addObject(char *objectname, Node *parentNodePtr, SmiSubid subid,
     objectPtr->listPtr			= NULL;
     objectPtr->flags			= flags;
     objectPtr->line			= parserPtr ? parserPtr->line : -1;
+    
+    objectPtr->uniquenessPtr            = NULL;
     					
     objectPtr->export.oidlen            = 0;     /* filled in by  */
     objectPtr->export.oid               = NULL;  /* second pass.  */
@@ -869,7 +898,7 @@ Object *addObject(char *objectname, Node *parentNodePtr, SmiSubid subid,
 	nodePtr->lastObjectPtr			      = objectPtr;
     }
     objectPtr->nodePtr				      = nodePtr;
-
+    
     return (objectPtr);
 }
 
@@ -951,7 +980,7 @@ Object *duplicateObject(Object *templatePtr, ObjectFlags flags,
 	nodePtr->lastObjectPtr->nextSameNodePtr       = objectPtr;
     nodePtr->lastObjectPtr			      = objectPtr;
     objectPtr->nodePtr				      = nodePtr;
-
+    
     return (objectPtr);
 }
 
@@ -1466,7 +1495,6 @@ void setObjectReference(Object *objectPtr, char *reference, Parser *parserPtr)
 }
 
 
-
 /*
  *----------------------------------------------------------------------
  *
@@ -1823,6 +1851,52 @@ void setObjectValue(Object *objectPtr, SmiValue *valuePtr)
 
 
 
+/*
+ *----------------------------------------------------------------------
+ *
+ * setObjectUniqueness --
+ *
+ *      Set the uniqueness entry of an object
+ *
+ * Results:
+ *	None.
+ *
+ * Side effects:
+ *      None.
+ *
+ *----------------------------------------------------------------------
+ */
+
+void setObjectUniqueness(Object *objectPtr, List *listPtr)
+{
+    objectPtr->uniquenessPtr = listPtr;
+}
+
+
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * setObjectInstallErrors --
+ *
+ *      Set the install errors entry of an object
+ *
+ * Results:
+ *	None.
+ *
+ * Side effects:
+ *      None.
+ *
+ *----------------------------------------------------------------------
+ */
+
+/*void setObjectInstallErrors(Object *objectPtr, List *listPtr)
+{
+    objectPtr->installErrorsPtr = listPtr;
+}*/
+
+
+
 /*
  *----------------------------------------------------------------------
  *
@@ -2357,7 +2431,8 @@ Type *duplicateType(Type *templatePtr, TypeFlags flags, Parser *parserPtr)
 Type *setTypeName(Type *typePtr, char *name)
 {
     Type              *type2Ptr;
-
+    List	      *listPtr;
+    
     if (typePtr->export.name) {
 	smiFree(typePtr->export.name);
     }
@@ -2406,6 +2481,18 @@ Type *setTypeName(Type *typePtr, char *name)
 	    type2Ptr->flags        = typePtr->flags;
 	    type2Ptr->line         = typePtr->line;
 
+	    /*
+	     * if it's an enum or bits type, we also have to adjust
+	     * the references from the named numbers back to the type.
+	     */
+	    if ((type2Ptr->export.basetype == SMI_BASETYPE_ENUM) ||
+		(type2Ptr->export.basetype == SMI_BASETYPE_BITS)) {
+		for (listPtr = type2Ptr->listPtr; listPtr;
+		     listPtr = listPtr->nextPtr) {
+		    ((NamedNumber *)(listPtr->ptr))->typePtr = type2Ptr;
+		}
+	    }
+	    
 	    smiFree(typePtr->export.name);
 	    smiFree(typePtr);
 
@@ -3273,16 +3360,19 @@ int smiInitData()
 	smiMalloc(sizeof(int));
     objectPtr->export.oidlen = objectPtr->nodePtr->oidlen = 1;
     objectPtr->nodePtr->oid[0] = 0;
+    objectPtr->export.nodekind = SMI_NODEKIND_NODE;
     objectPtr = addObject(smiStrdup("iso"), smiHandle->rootNodePtr, 1, 0, &parser);
     objectPtr->export.oid = objectPtr->nodePtr->oid =
 	smiMalloc(sizeof(int));
     objectPtr->export.oidlen = objectPtr->nodePtr->oidlen = 1;
     objectPtr->nodePtr->oid[0] = 1;
+    objectPtr->export.nodekind = SMI_NODEKIND_NODE;
     objectPtr = addObject(smiStrdup("joint-iso-ccitt"), smiHandle->rootNodePtr, 2, 0, &parser);
     objectPtr->export.oid = objectPtr->nodePtr->oid =
 	smiMalloc(sizeof(int));
     objectPtr->export.oidlen = objectPtr->nodePtr->oidlen = 1;
     objectPtr->nodePtr->oid[0] = 2;
+    objectPtr->export.nodekind = SMI_NODEKIND_NODE;
     
     
     smiHandle->typeOctetStringPtr =
@@ -3350,6 +3440,13 @@ static void freeNodeTree(Node *rootPtr)
 	smiFree(nodePtr->oid);
 	smiFree(nodePtr);
     }
+    rootPtr->firstChildPtr = NULL;
+    rootPtr->lastChildPtr = NULL;
+    rootPtr->firstObjectPtr = NULL;
+    rootPtr->lastObjectPtr = NULL;
+    rootPtr->nextPtr = NULL;
+    rootPtr->prevPtr = NULL;
+    rootPtr->parentPtr = NULL;
 }
 
 
@@ -3387,7 +3484,12 @@ void smiFreeData()
 	smiFree(viewPtr);
     }
 
-    for (modulePtr = smiHandle->firstModulePtr; modulePtr; modulePtr = nextModulePtr) {
+    /*
+     * In this first module loop we remove each module's imports,
+     * revisions, macros, and objects.
+     */
+    for (modulePtr = smiHandle->firstModulePtr; modulePtr;
+	 modulePtr = nextModulePtr) {
 	nextModulePtr = modulePtr->nextPtr;
 
 	for (importPtr = modulePtr->firstImportPtr; importPtr;
@@ -3455,7 +3557,17 @@ void smiFreeData()
 	    }
 	    smiFree(objectPtr);
 	}
-	
+    }
+
+    /*
+     * In this second module loop we remove each module's types
+     * and the modules themselves. This separation is required, because
+     * we reference some types of foreign modules in the first loop.
+     */
+    for (modulePtr = smiHandle->firstModulePtr; modulePtr;
+	 modulePtr = nextModulePtr) {
+	nextModulePtr = modulePtr->nextPtr;
+
 	for (typePtr = modulePtr->firstTypePtr; typePtr;
 	     typePtr = nextTypePtr) {
 	    nextTypePtr = typePtr->nextPtr;
@@ -3520,7 +3632,7 @@ void smiFreeData()
  *----------------------------------------------------------------------
  */
 
-Module *loadModule(const char *modulename)
+Module *loadModule(const char *modulename, Parser *parserPtr)
 {
     Parser	    parser;
     char	    *path = NULL, *dir, *smipath;
@@ -3545,28 +3657,32 @@ Module *loadModule(const char *modulename)
 	sep[0] = PATH_SEPARATOR; sep[1] = 0;
 	for (dir = strtok(smipath, sep);
 	     dir; dir = strtok(NULL, sep)) {
-	    path = smiMalloc(strlen(dir)+strlen(modulename)+8);
-	    sprintf(path, "%s%c%s", dir, DIR_SEPARATOR, modulename);
+	    smiAsprintf(&path, "%s%c%s", dir, DIR_SEPARATOR, modulename);
 	    if (! access(path, R_OK)) {
 		break;
 	    }
-	    sprintf(path, "%s%c%s.my", dir, DIR_SEPARATOR, modulename);
+	    smiFree(path);
+	    smiAsprintf(&path, "%s%c%s.my", dir, DIR_SEPARATOR, modulename);
 	    if (! access(path, R_OK)) {
 		break;
 	    }
-	    sprintf(path, "%s%c%s.smiv2", dir, DIR_SEPARATOR, modulename);
+	    smiFree(path);
+	    smiAsprintf(&path, "%s%c%s.smiv2", dir, DIR_SEPARATOR, modulename);
 	    if (! access(path, R_OK)) {
 		break;
 	    }
-	    sprintf(path, "%s%c%s.sming", dir, DIR_SEPARATOR, modulename);
+	    smiFree(path);
+	    smiAsprintf(&path, "%s%c%s.sming", dir, DIR_SEPARATOR, modulename);
 	    if (! access(path, R_OK)) {
 		break;
 	    }
-	    sprintf(path, "%s%c%s.mib", dir, DIR_SEPARATOR, modulename);
+	    smiFree(path);
+	    smiAsprintf(&path, "%s%c%s.mib", dir, DIR_SEPARATOR, modulename);
 	    if (! access(path, R_OK)) {
 		break;
 	    }
-	    sprintf(path, "%s%c%s.txt", dir, DIR_SEPARATOR, modulename);
+	    smiFree(path);
+	    smiAsprintf(&path, "%s%c%s.txt", dir, DIR_SEPARATOR, modulename);
 	    if (! access(path, R_OK)) {
 		break;
 	    }
@@ -3581,19 +3697,17 @@ Module *loadModule(const char *modulename)
 	path = smiStrdup(modulename);
     }
 
-#if !defined(_MSC_VER)
+#if !defined(_MSC_VER) && !defined(__MINGW32__)
     if (!path && smiHandle->cache && smiHandle->cacheProg) {
 	/* Not found in the path; now try to fetch & cache the module. */
 	int  pid;
 	char *argv[4];
 	char *cmd;
 	int  status;
-	path = smiMalloc(strlen(smiHandle->cache) + strlen(modulename) + 2);
-	sprintf(path, "%s%c%s", smiHandle->cache, DIR_SEPARATOR, modulename);
+	smiAsprintf(&path, "%s%c%s",
+		    smiHandle->cache, DIR_SEPARATOR, modulename);
 	if (access(path, R_OK)) {
-	    cmd = smiMalloc(strlen(smiHandle->cacheProg) +
-			    strlen(modulename) + 2);
-	    sprintf(cmd, "%s %s", smiHandle->cacheProg, modulename);
+	    smiAsprintf(&cmd, "%s %s", smiHandle->cacheProg, modulename);
 	    pid = fork();
 	    if (pid != -1) {
 		if (!pid) {
@@ -3613,7 +3727,7 @@ Module *loadModule(const char *modulename)
 #endif
     
     if (!path) {
-	smiPrintError(NULL, ERR_MODULE_NOT_FOUND, modulename);
+	smiPrintError(parserPtr, ERR_MODULE_NOT_FOUND, modulename);
 	return NULL;
     }
 
@@ -3626,7 +3740,7 @@ Module *loadModule(const char *modulename)
 
     file = fopen(path, "r");
     if (! file) {
-	smiPrintError(NULL, ERR_OPENING_INPUTFILE, path, strerror(errno));
+	smiPrintError(parserPtr, ERR_OPENING_INPUTFILE, path, strerror(errno));
 	smiFree(path);
 	return NULL;
     }
@@ -3638,7 +3752,7 @@ Module *loadModule(const char *modulename)
 	    sming = 1;
 	    break;
 	} else if (c == EOF || ! isspace(c)) {
-	    smiPrintError(NULL, ERR_ILLEGAL_INPUTFILE, path);
+	    smiPrintError(parserPtr, ERR_ILLEGAL_INPUTFILE, path);
 	    smiFree(path);
 	    fclose(file);
 	    return NULL;
@@ -3659,13 +3773,14 @@ Module *loadModule(const char *modulename)
 	smiDepth++;
 	parser.line			= 1;
 	smiparse((void *)&parser);
+	freeNodeTree(smiHandle->pendingNodePtr);
 	smiLeaveLexRecursion();
 	smiDepth--;
 	fclose(parser.file);
 	smiFree(path);
 	return parser.modulePtr;
 #else
-	smiPrintError(NULL, ERR_SMI_NOT_SUPPORTED, path);
+	smiPrintError(parserPtr, ERR_SMI_NOT_SUPPORTED, path);
 	smiFree(path);
         fclose(file);
 	return NULL;
@@ -3685,13 +3800,14 @@ Module *loadModule(const char *modulename)
 	smiDepth++;
 	parser.line			= 1;
 	smingparse((void *)&parser);
+	freeNodeTree(smiHandle->pendingNodePtr);
 	smingLeaveLexRecursion();
 	smiDepth--;
 	fclose(parser.file);
 	smiFree(path);
 	return parser.modulePtr;
 #else
-	smiPrintError(NULL, ERR_SMING_NOT_SUPPORTED, path);
+	smiPrintError(parserPtr, ERR_SMING_NOT_SUPPORTED, path);
 	smiFree(path);
         fclose(file);
 	return NULL;
