@@ -8,16 +8,19 @@
  * See the file "COPYING" for information on usage and redistribution
  * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
  *
- * @(#) $Id: util.c,v 1.28 2002/01/09 09:09:00 strauss Exp $
+ * @(#) $Id: util.c,v 1.33 2002/11/13 13:15:03 strauss Exp $
  */
 
 #include <config.h>
 
+#include <stdio.h>
+#include <stdarg.h>
 #include <stdlib.h>
 #include <string.h>
 #include <time.h>
 
 #include "util.h"
+#include "snprintf.h"
 
 #ifdef HAVE_DMALLOC_H
 #include <dmalloc.h>
@@ -154,15 +157,14 @@ time_t timegm(struct tm *tm)
     tz = getenv("TZ");
     if (tz) {
 	tofree = s;
-	s = smiMalloc(strlen(tz)+4);
-	sprintf(s, "TZ=%s", tz);
+	smiAsprintf(&s, "TZ=%s", tz);
     }
     putenv("TZ=NULL");
     t = mktime(tm);
     if (tz) {
 	putenv(s);
     } else {
-	putenv("TZ");
+	putenv("TZ=");
     }
     if (tofree) smiFree(tofree);
 
@@ -175,6 +177,10 @@ time_t timegm(struct tm *tm)
 int smiTypeDerivedFrom(Type *typePtr, Type *parentTypePtr)
 {
     Type *t;
+
+    if ((!typePtr) || (!parentTypePtr)) {
+	return 0;
+    }
     
     for (t = typePtr; t != NULL; t = t->parentPtr) {
 	if (parentTypePtr == t) {
