@@ -12,7 +12,7 @@
  * See the file "COPYING" for information on usage and redistribution
  * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
  *
- * @(#) $Id: dump-corba.c,v 1.41 2001/11/08 07:37:31 schoenw Exp $
+ * @(#) $Id: dump-corba.c,v 1.45 2002/10/30 09:17:37 schoenw Exp $
  */
 
 #include <config.h>
@@ -451,7 +451,7 @@ static char *getValueString(SmiValue *valuePtr, SmiType *typePtr)
             sprintf(s, "'%*s'H", 2 * valuePtr->len, "");
             for (i=0; i < valuePtr->len; i++) {
                 sprintf(ss, "%02x", valuePtr->value.ptr[i]);
-                strncpy(&s[2+2*i], ss, 2);
+                strncpy(&s[1+2*i], ss, 2);
             }
 	}
 	break;
@@ -529,24 +529,51 @@ static void createImportList(SmiModule *smiModule)
     SmiNode     *smiNode;
     SmiType     *smiType;
     SmiModule   *smiTypeModule;
-    SmiNodekind kind = SMI_NODEKIND_SCALAR | SMI_NODEKIND_COLUMN;
+    SmiElement  *smiElement;
+    SmiNode     *smiNodeIndex;
+    SmiNodekind kind = SMI_NODEKIND_SCALAR
+	 | SMI_NODEKIND_COLUMN | SMI_NODEKIND_ROW;
     
     for (smiNode = smiGetFirstNode(smiModule, kind);
 	 smiNode;
 	 smiNode = smiGetNextNode(smiNode, kind)) {
 
-	smiType = smiGetNodeType(smiNode);
-	if (smiType) {
-	    smiTypeModule = smiGetTypeModule(smiType);
-	    if (smiTypeModule &&
-		strcmp(smiTypeModule->name, smiModule->name)) {
-		if (strlen(smiTypeModule->name)) {
-		    addImport(smiTypeModule->name, smiType->name);
-		}
-	    }
-	    if (smiType->basetype == SMI_BASETYPE_INTEGER32) {
-		addImport("SNMPv2-SMI", "Integer32");
-	    }
+	 switch (smiNode->nodekind) {
+	 case SMI_NODEKIND_ROW:
+	      for (smiElement = smiGetFirstElement(smiNode); smiElement;
+		   smiElement = smiGetNextElement(smiElement)) {
+		   smiNodeIndex = smiGetElementNode(smiElement);
+		  smiType = smiGetNodeType(smiNodeIndex);
+		  if (smiType) {
+		      smiTypeModule = smiGetTypeModule(smiType);
+		      if (smiTypeModule &&
+			  strcmp(smiTypeModule->name, smiModule->name)) {
+			  if (strlen(smiTypeModule->name)) {
+			      addImport(smiTypeModule->name, smiType->name);
+			  }
+		      }
+		      if (smiType->basetype == SMI_BASETYPE_INTEGER32) {
+			   addImport("SNMPv2-SMI", "Integer32");
+		      }
+		  }
+	      }
+	      break;
+	 case SMI_NODEKIND_SCALAR:
+	 case SMI_NODEKIND_COLUMN:
+	      smiType = smiGetNodeType(smiNode);
+	      if (smiType) {
+		   smiTypeModule = smiGetTypeModule(smiType);
+		   if (smiTypeModule &&
+		       strcmp(smiTypeModule->name, smiModule->name)) {
+			if (strlen(smiTypeModule->name)) {
+			     addImport(smiTypeModule->name, smiType->name);
+			}
+		   }
+		   if (smiType->basetype == SMI_BASETYPE_INTEGER32) {
+			addImport("SNMPv2-SMI", "Integer32");
+		   }
+	      }
+	      break;
 	}
     }
 }
@@ -1078,8 +1105,10 @@ static void fprintInterfaces(FILE *f, SmiModule *smiModule)
 static void fprintConstructor(FILE *f, SmiNode *smiNode)
 {
     SmiNode *childNode;
+    SmiNode *indexNode;
     SmiType *smiType;
     SmiModule *smiModule;
+    SmiElement *smiElement = NULL;
     char    *idlNodeName;
     char    *idlChildNodeName, *idlChildTypeName;
     int	    cnt = 0;
@@ -1091,26 +1120,64 @@ static void fprintConstructor(FILE *f, SmiNode *smiNode)
     fprintSegment(f, 2*INDENT, "", 0);
     fprint(f, "%s create_%s (\n", idlNodeName, idlNodeName);
 
-    /* XXX I think we need to include INDEX columns here XXX */
+    /* First include the INDEXes as parameters to allow row creation
+       for rows with not-accesible index objects. */
+
+    if (smiNode->indexkind == SMI_INDEX_INDEX
+	|| smiNode->indexkind == SMI_INDEX_REORDER) {
+	 for (smiElement = smiGetFirstElement(smiNode);
+	      smiElement; smiElement = smiGetNextElement(smiElement)) {
+	      cnt++;
+	      indexNode = smiGetElementNode(smiElement);
+	      idlChildNodeName =
+		   getIdlNodeName(smiGetNodeModule(indexNode)->name,
+				  indexNode->name);
+	      smiType = smiGetNodeType(indexNode);
+	      idlChildTypeName = getIdlAnyTypeName(indexNode, smiType);
+	      if (cnt > 1) {
+		   fprint(f, ",\n");
+	      }
+	      fprintSegment(f, 3*INDENT, "in ", 0);
+	      fprint(f, "%s %s", idlChildTypeName, idlChildNodeName);
+	 }
+    }
+    
     for (childNode = smiGetFirstChildNode(smiNode);
 	 childNode;
 	 childNode = smiGetNextChildNode(childNode)) {
+	
 	if ((childNode->nodekind == SMI_NODEKIND_SCALAR
 	     || childNode->nodekind == SMI_NODEKIND_COLUMN)
 	    && current(childNode->status)
 	    && childNode->access == SMI_ACCESS_READ_WRITE) {
 
-	    cnt++;
-	    idlChildNodeName =
-		getIdlNodeName(smiGetNodeModule(childNode)->name,
-			       childNode->name);
-	    smiType = smiGetNodeType(childNode);
-	    idlChildTypeName = getIdlAnyTypeName(childNode, smiType);
-	    if (cnt > 1) {
-		fprint(f, ",\n");
+	    /* Test if this column is already used as parameter
+	       because it is an INDEX of the row. */
+
+	    if (childNode->nodekind == SMI_NODEKIND_SCALAR
+		|| childNode->nodekind == SMI_NODEKIND_COLUMN) {
+		for (smiElement = smiGetFirstElement(smiNode);
+		    smiElement; smiElement = smiGetNextElement(smiElement)) {
+		    indexNode = smiGetElementNode(smiElement);
+		    if (indexNode == childNode) {
+			break;
+		    }
+		}
+	    }
+
+	    if (! smiElement) {
+		cnt++;
+		idlChildNodeName =
+		    getIdlNodeName(smiGetNodeModule(childNode)->name,
+				   childNode->name);
+		smiType = smiGetNodeType(childNode);
+		idlChildTypeName = getIdlAnyTypeName(childNode, smiType);
+		if (cnt > 1) {
+		    fprint(f, ",\n");
+		}
+		fprintSegment(f, 3*INDENT, "in ", 0);
+		fprint(f, "%s %s", idlChildTypeName, idlChildNodeName);
 	    }
-	    fprintSegment(f, 3*INDENT, "in ", 0);
-	    fprint(f, "%s %s", idlChildTypeName, idlChildNodeName);
 	}
     }
     fprint(f, "\n");
@@ -1464,6 +1531,11 @@ static void dumpIdl(SmiModule *smiModule)
     dictFree(&idlTypeNameList);
     dictFree(&idlVBTypeNameList);
 
+    if (fflush(f) || ferror(f)) {
+	perror("smidump: write error");
+	exit(1);
+    }
+
     fclose(f);
 }
 
@@ -1567,6 +1639,12 @@ static void dumpOid(SmiModule *smiModule)
     }
 
     dictFree(&idlModuleNameList);
+
+    if (fflush(f) || ferror(f)) {
+	perror("smidump: write error");
+	exit(1);
+    }
+
     fclose(f);
 }
 
