@@ -10,7 +10,7 @@
  * See the file "COPYING" for information on usage and redistribution
  * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
  *
- * @(#) $Id: dump-python.c,v 1.11 2001/11/08 07:37:31 schoenw Exp $
+ * @(#) $Id: dump-python.c,v 1.12 2002/10/30 09:17:37 schoenw Exp $
  */
 
 /*
@@ -1046,6 +1046,11 @@ static void dumpPython(int modc, SmiModule **modv, int flags, char *output)
 	fprint(f, "}\n");
     }
 
+    if (fflush(f) || ferror(f)) {
+	perror("smidump: write error");
+	exit(1);
+    }
+
     if (output) {
 	fclose(f);
     } 
