@@ -8,7 +8,7 @@
  * See the file "COPYING" for information on usage and redistribution
  * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
  *
- * @(#) $Id: dump-smi.c,v 1.75 2002/01/09 12:02:03 schoenw Exp $
+ * @(#) $Id: dump-smi.c,v 1.84 2002/12/17 17:50:03 schoenw Exp $
  */
 
 #include <config.h>
@@ -24,9 +24,15 @@
 #include "smidump.h"
 
 
+
+extern int smiAsprintf(char **strp, const char *format, ...);
+
+
+
 #define  INDENT		4    /* indent factor */
 #define  INDENTVALUE	16   /* column to start values, except multiline */
 #define  INDENTTEXTS	 9   /* column to start multiline texts */
+#define  INDENTTEXTS2	15   /* column to start indented multiline texts */
 #define  INDENTMAX	72   /* max column to fill, break lines otherwise */
 
 
@@ -52,6 +58,15 @@ static char *convertTypev2[] = {
     NULL,		 "Gauge",	       NULL,	   "Gauge32",
     NULL, NULL, NULL, NULL };
 
+static char *convertTypePIBtoMIB[] = {
+    NULL,		 "Unsigned64",	       NULL,	   "IBRUnsigned64",
+    NULL,		 "Integer64",	       NULL,	   "IBRInteger64",
+/* XXX This breaks SMIv2 -> SMIv? compilation such that the IMPORT statement
+   for SnmpAdminString is removed. I've remove this as a quick fix, but I'm
+   not sure what to do for PIB compilation here.
+    NULL,		 "SnmpAdminString",    NULL,	   "OCTET STRING",
+*/
+    NULL, NULL, NULL, NULL };
 
 static char *convertImportv2[] = {
     "IRTF-NMRG-SMING-TYPES", "IpAddress",       "SNMPv2-SMI", "IpAddress",
@@ -91,6 +106,9 @@ static char *convertImportv2[] = {
     "RFC1155-SMI", "Opaque",        "SNMPv2-SMI", "Opaque",
     "RFC1213-MIB", "mib-2",         "SNMPv2-SMI", "mib-2",    
     "RFC1213-MIB", "DisplayString", "SNMPv2-TC",  "DisplayString",    
+
+/* import types for PIB to MIB conversion */
+
     NULL, NULL, NULL, NULL };
 
 static char *convertImportv1[] = {
@@ -125,7 +143,7 @@ static char *convertImportv1[] = {
 static int current_column = 0;
 static int smiv1 = 0;
 static int silent = 0;
-
+static int pibtomib = 0;
 
 
 typedef struct Import {
@@ -146,11 +164,15 @@ static int invalidType(SmiBasetype basetype)
 	    || (basetype == SMI_BASETYPE_FLOAT32)
 	    || (basetype == SMI_BASETYPE_FLOAT64)
 	    || (basetype == SMI_BASETYPE_FLOAT128);
-    } else {
+    } else if (!pibtomib) {
 	return (basetype == SMI_BASETYPE_INTEGER64)
 	    || (basetype == SMI_BASETYPE_FLOAT32)
 	    || (basetype == SMI_BASETYPE_FLOAT64)
 	    || (basetype == SMI_BASETYPE_FLOAT128);
+    } else {
+	return (basetype == SMI_BASETYPE_FLOAT32)
+	    || (basetype == SMI_BASETYPE_FLOAT64)
+	    || (basetype == SMI_BASETYPE_FLOAT128);
     }
 }
 
@@ -193,12 +215,22 @@ static char *getAccessString(SmiAccess access, int create)
 	if (create && (access == SMI_ACCESS_READ_WRITE)) {
 	    return "read-create";
 	} else {
-	    return
-	    (access == SMI_ACCESS_NOT_ACCESSIBLE) ? "not-accessible" :
-	    (access == SMI_ACCESS_NOTIFY)	  ? "accessible-for-notify" :
-	    (access == SMI_ACCESS_READ_ONLY)      ? "read-only" :
-	    (access == SMI_ACCESS_READ_WRITE)     ? "read-write" :
-						    "<unknown>";
+            if (pibtomib)
+                return
+	        (access == SMI_ACCESS_NOT_ACCESSIBLE) ? "not-accessible" :
+	        (access == SMI_ACCESS_NOTIFY)	      ? "read-only" :
+	        (access == SMI_ACCESS_READ_ONLY)      ? "read-only" :
+	        (access == SMI_ACCESS_READ_WRITE)     ? "read-write" :
+	        (access == SMI_ACCESS_INSTALL)        ? "read-create" :
+	        (access == SMI_ACCESS_INSTALL_NOTIFY) ? "read-create" :
+						        "<unknown>";
+            else
+	        return
+	        (access == SMI_ACCESS_NOT_ACCESSIBLE) ? "not-accessible" :
+	        (access == SMI_ACCESS_NOTIFY)	  ? "accessible-for-notify" :
+	        (access == SMI_ACCESS_READ_ONLY)      ? "read-only" :
+	        (access == SMI_ACCESS_READ_WRITE)     ? "read-write" :
+						        "<unknown>";
 	}
     }
 }
@@ -207,13 +239,15 @@ static char *getAccessString(SmiAccess access, int create)
 
 static char *getTimeString(time_t t)
 {
-    static char   s[27];
+    static char   *s = NULL;
     struct tm	  *tm;
 
+    if (s) xfree(s);
+    
     tm = gmtime(&t);
-    sprintf(s, "%04d%02d%02d%02d%02dZ",
-	    tm->tm_year + 1900, tm->tm_mon + 1, tm->tm_mday,
-	    tm->tm_hour, tm->tm_min);
+    smiAsprintf(&s, "%04d%02d%02d%02d%02dZ",
+		tm->tm_year + 1900, tm->tm_mon + 1, tm->tm_mday,
+		tm->tm_hour, tm->tm_min);
     return s;
 }
 
@@ -233,6 +267,14 @@ static char *getTypeString(SmiBasetype basetype, SmiType *smiType)
     if (typeName &&
 	(basetype != SMI_BASETYPE_ENUM) &&
 	(basetype != SMI_BASETYPE_BITS)) {
+        if (pibtomib)
+	    for(i=0; convertTypePIBtoMIB[i+1]; i += 4) {
+	        if ((!strcmp(typeName, convertTypePIBtoMIB[i+1])) &&
+		    ((!typeModule) || (!convertTypePIBtoMIB[i]) ||
+		     (!strcmp(typeModule, convertTypePIBtoMIB[i])))) {
+		    return convertTypePIBtoMIB[i+3];
+	        }
+	    }
 	for(i=0; convertType[i+1]; i += 4) {
 	    if ((!strcmp(typeName, convertType[i+1])) &&
 		((!typeModule) || (!convertType[i]) ||
@@ -427,18 +469,66 @@ static void createImportList(SmiModule *smiModule)
     SmiImport   *smiImport;
     SmiModule	*smiModule2;
     
+    if (pibtomib) {
+        addImport("TUBS-SMI", "ibrpibtomib");
+	for(smiType = smiGetFirstType(smiModule);
+	    smiType; smiType = smiGetNextType(smiType)) {
+	    if (smiType->basetype == SMI_BASETYPE_UNSIGNED32)
+	        addImport("SNMPv2-SMI", "Unsigned32"); 
+            else if (smiType->basetype == SMI_BASETYPE_UNSIGNED64)
+	        addImport("TUBS-SMI", "IBRUnsigned64"); 
+            else if (smiType->basetype == SMI_BASETYPE_INTEGER64)
+	        addImport("TUBS-SMI", "IBRInteger64"); 
+        }
+        if (smiGetFirstNode(smiModule, SMI_NODEKIND_TABLE) != NULL)
+            addImport("SNMPv2-TC", "RowStatus");
+        for(smiNode = smiGetFirstNode(smiModule, SMI_NODEKIND_ROW);
+            smiNode; smiNode = smiGetNextNode(smiNode, SMI_NODEKIND_ROW)) {
+            if (smiNode->indexkind == SMI_INDEX_SPARSE) {
+                SmiNode *smiNode2 = smiGetRelatedNode(smiNode);
+                while (smiNode2 && (smiNode2->indexkind == SMI_INDEX_SPARSE))
+                    smiNode2 = smiGetRelatedNode(smiNode2);
+                if (smiNode2) {
+                    smiModule2 = smiGetNodeModule(smiNode2);
+                    if (smiModule != smiModule2)
+                        addImport(smiModule2->name, smiNode2->name);
+                }
+            }
+        }
+    }
+        
     for(smiNode = smiGetFirstNode(smiModule, kind); smiNode;
 	smiNode = smiGetNextNode(smiNode, kind)) {
 	smiType = smiGetNodeType(smiNode);
+	if (smiType && (smiType->decl == SMI_DECL_IMPLICIT_TYPE)) {
+	    smiType = smiGetParentType(smiType);
+	}
 	if (smiType) {
 	    smiModule2 = smiGetTypeModule(smiType);
 	    if (smiModule2 && (smiModule2 != smiModule)) {
-		if (strlen(smiModule2->name) && smiType->name) {
-		    addImport(smiModule2->name, smiType->name);
+                size_t len = strlen(smiModule2->name);
+		if (len && smiType->name) {
+                    if (pibtomib && (len > 4) &&
+                        !strcmp(&smiModule2->name[len - 4], "-PIB")) {
+                        char *newName = xmalloc(len + 5);
+                        strcpy(newName, smiModule2->name);
+                        strcat(newName, "-MIB");
+                        addImport(newName, smiType->name);
+                    } else
+		        addImport(smiModule2->name, smiType->name);
 		}
 	    }
 	}
 
+	if (pibtomib) {
+            if (smiType && (smiType->basetype == SMI_BASETYPE_UNSIGNED32))
+	        addImport("SNMPv2-SMI", "Unsigned32");
+            if (smiType && (smiType->basetype == SMI_BASETYPE_UNSIGNED64))
+	        addImport("TUBS-SMI", "IBRUnsigned64");
+            if (smiType && (smiType->basetype == SMI_BASETYPE_INTEGER64))
+	        addImport("TUBS-SMI", "IBRInteger64");
+	}
+
 	if (! smiv1 &&
 	    smiType && smiType->basetype == SMI_BASETYPE_INTEGER32) {
 	    addImport("SNMPv2-SMI", "Integer32");
@@ -626,6 +716,28 @@ static void fprintMultilineString(FILE *f, const char *s, const int comment)
 
 
 
+static void fprintMultilineString2(FILE *f, const char *s, const int comment)
+{
+    int i, len;
+    
+    fprintSegment(f, INDENTTEXTS2 - 1, "\"", 0, comment);
+    if (s) {
+	len = strlen(s);
+	for (i=0; i < len; i++) {
+	    putc(s[i], f);
+	    current_column++;
+	    if (s[i] == '\n') {
+		current_column = 0;
+		fprintSegment(f, INDENTTEXTS2, "", 0, comment);
+	    }
+	}
+    }
+    putc('\"', f);
+    current_column++;
+}
+
+
+
 static char *getValueString(SmiValue *valuePtr, SmiType *typePtr)
 {
     static char    s[100];
@@ -734,9 +846,14 @@ static void fprintSubtype(FILE *f, SmiType *smiType, const int comment)
     SmiRange       *range;
     SmiNamedNumber *nn;
     char	   s[100];
-    int		   i, c;
+    int		   i, c = comment;
 
-    c = comment || (smiv1 && smiType->basetype == SMI_BASETYPE_BITS);
+    if (pibtomib && ((smiType->basetype == SMI_BASETYPE_UNSIGNED64) ||
+        (smiType->basetype == SMI_BASETYPE_INTEGER64))) {
+        c = 1;
+        fprint(f, " -- commented out by smidump\n-- ");
+    }
+    c = c || (smiv1 && smiType->basetype == SMI_BASETYPE_BITS);
     
     if ((smiType->basetype == SMI_BASETYPE_ENUM) ||
 	(smiType->basetype == SMI_BASETYPE_BITS)) {
@@ -795,14 +912,18 @@ static void fprintSubtype(FILE *f, SmiType *smiType, const int comment)
 
 static void fprintIndex(FILE *f, SmiNode *indexNode, const int comment)
 {
-    SmiElement *smiElement;
+    SmiElement *smiElement, *smiFirstElement;
     int        n, j;
 
-    fprintSegment(f, INDENT, "INDEX", INDENTVALUE, comment);
-    fprint(f, "{ ");
     for (n = 0, smiElement = smiGetFirstElement(indexNode); smiElement;
 	 n++, smiElement = smiGetNextElement(smiElement));
-    for (j = 0, smiElement = smiGetFirstElement(indexNode); smiElement;
+    if (pibtomib && (n > 1))
+        smiFirstElement = smiGetNextElement(smiGetFirstElement(indexNode));
+    else
+        smiFirstElement = smiGetFirstElement(indexNode);
+    fprintSegment(f, INDENT, "INDEX", INDENTVALUE, comment);
+    fprint(f, "{ ");
+    for (j = 0, smiElement = smiFirstElement; smiElement;
 	 j++, smiElement = smiGetNextElement(smiElement)) {
 	if (j) {
 	    fprint(f, ", ");
@@ -946,7 +1067,11 @@ static void fprintModuleIdentity(FILE *f, SmiModule *smiModule)
 	    
 	    if (smiNode) {
 		fprintSegment(f, INDENT, "::= ", 0, smiv1);
-		fprint(f, "{ %s }\n\n", getOidString(smiNode, 0));
+                if (!pibtomib)
+		    fprint(f, "{ %s }\n\n", getOidString(smiNode, 0));
+                else
+                    fprint(f, "{ ibrpibtomib %d }\n\n",
+                           smiNode->oid[smiNode->oidlen - 1]);
 	    }
 	    /* TODO: else error */
 	    
@@ -1050,13 +1175,30 @@ static void fprintTextualConventions(FILE *f, SmiModule *smiModule)
 }
 
 
+static int isInIndex(SmiNode *node, SmiNode *parentNode)
+{
+    SmiElement *element;
+    
+    if ((parentNode->indexkind != SMI_INDEX_INDEX) &&
+        (parentNode->indexkind != SMI_INDEX_EXPAND))
+        return 0;
+    
+    for (element = smiGetNextElement(smiGetFirstElement(parentNode)); element;
+         element = smiGetNextElement(element))
+        if (smiGetElementNode(element)->name &&
+            node->name &&
+            !strcmp(smiGetElementNode(element)->name, node->name))
+            return 1;
+    return 0;
+}
+
 
 static void fprintObjects(FILE *f, SmiModule *smiModule)
 {
     SmiNode	 *smiNode, *rowNode, *colNode, *smiParentNode, *relatedNode;
     SmiType	 *smiType;
     SmiNodekind  nodekinds;
-    int		 i, invalid, create, assignement, indentsequence;
+    int		 i, invalid, create, assignement, indentsequence, addrowstatus;
     
     nodekinds =  SMI_NODEKIND_NODE | SMI_NODEKIND_TABLE |
 	SMI_NODEKIND_ROW | SMI_NODEKIND_COLUMN | SMI_NODEKIND_SCALAR |
@@ -1169,9 +1311,21 @@ static void fprintObjects(FILE *f, SmiModule *smiModule)
 	    if (smiv1) {
 		fprintSegment(f, INDENT, "ACCESS", INDENTVALUE, invalid);
 	    } else {
-		fprintSegment(f, INDENT, "MAX-ACCESS", INDENTVALUE, 0);
+ 		fprintSegment(f, INDENT, "MAX-ACCESS", INDENTVALUE, 0);
 	    }
-	    fprint(f, "%s\n", getAccessString(smiNode->access, create));
+            if (pibtomib) {
+                if ((smiNode->nodekind == SMI_NODEKIND_TABLE) ||
+                    (smiNode->nodekind == SMI_NODEKIND_ROW))
+                    fprint(f, "not-accessible\n");
+                else if ((smiNode->nodekind == SMI_NODEKIND_COLUMN) &&
+                         isInIndex(smiNode, smiParentNode))
+	            fprint(f, "not-accessible\n");
+                else if (smiNode->nodekind == SMI_NODEKIND_COLUMN)
+                    fprint(f, "read-create\n");
+                else
+	            fprint(f, "%s\n", getAccessString(smiNode->access, create));
+            } else
+	        fprint(f, "%s\n", getAccessString(smiNode->access, create));
 	}
 
 	if (! assignement) {
@@ -1221,6 +1375,9 @@ static void fprintObjects(FILE *f, SmiModule *smiModule)
 	    break;
 	case SMI_INDEX_SPARSE:
 	    if (! invalid) {
+                while (relatedNode &&
+                       (relatedNode->indexkind == SMI_INDEX_SPARSE))
+                    relatedNode = smiGetRelatedNode(relatedNode);
 		if (relatedNode) {
 		    fprintIndex(f, relatedNode, invalid);
 		}
@@ -1241,7 +1398,10 @@ static void fprintObjects(FILE *f, SmiModule *smiModule)
 	fprint(f, "{ %s }\n\n", getOidString(smiNode, 0));
 
 	smiType = smiGetNodeType(smiNode);
+        addrowstatus = 0;
 	if (smiNode->nodekind == SMI_NODEKIND_ROW) {
+            if (pibtomib)
+                addrowstatus = 1;
 	    if (smiType) {
 		fprint(f, "%s ::= SEQUENCE {", smiType->name);
 	    } else {
@@ -1264,6 +1424,12 @@ static void fprintObjects(FILE *f, SmiModule *smiModule)
 		    relatedNode = colNode;
 		}
 	    }
+            if (pibtomib) {
+                int len = strlen(smiParentNode->name) + 9;
+                if (len > 64)
+                  len = 64;
+                indentsequence = len;
+            }
 	    if (relatedNode) relatedNode = smiGetNextChildNode(relatedNode);
 	    indentsequence = (2*INDENT + indentsequence + 1) / INDENT * INDENT;
 	    /* TODO: non-local name? */
@@ -1273,7 +1439,7 @@ static void fprintObjects(FILE *f, SmiModule *smiModule)
 		if (! invalid || ! silent) {
 		    if (i && (relatedNode != colNode)) {
 			fprint(f, ",");
-		    }
+		    } 
 		    fprint(f, "\n");
 		}
 		
@@ -1295,7 +1461,42 @@ static void fprintObjects(FILE *f, SmiModule *smiModule)
 		}
 		i++;
 	    }
-	    fprint(f, "\n}\n\n");
+            if (pibtomib) {
+                size_t len = strlen(smiParentNode->name);
+                int maxid;
+                char *rowStatus = xmalloc(len + 10);
+                strcpy(rowStatus, smiParentNode->name);
+                if (len > 55)
+                  len = 55;
+                strcpy(&rowStatus[len], "RowStatus");
+                fprint(f, ",\n");
+                fprintSegment(f, INDENT, rowStatus, indentsequence, 0);
+                fprint(f, "RowStatus\n}\n\n");
+                
+                fprint(f, "%s OBJECT-TYPE\n", rowStatus);
+	        fprintSegment(f, INDENT, "SYNTAX", INDENTVALUE, 0);
+                fprint(f, "RowStatus\n");
+                fprintSegment(f, INDENT, "MAX-ACCESS", INDENTVALUE, 0);
+                fprint(f, "read-create\n");
+                fprintSegment(f, INDENT, "STATUS", INDENTVALUE, 0);
+                fprint(f, "current\n");
+                fprintSegment(f, INDENT, "DESCRIPTION", INDENTVALUE, 0);
+                fprint(f, "\n");
+		fprintMultilineString(f, "Added by smidump for automatic " \
+                                      "PIB to MIB conversion.", 0);
+                fprint(f, "\n");
+	        fprintSegment(f, INDENT, "::= ", 0, 0);
+                for (maxid = 0, colNode = smiGetFirstChildNode(smiNode);
+                     colNode; colNode = smiGetNextChildNode(colNode))
+                     if (colNode->oidlen &&
+                         (colNode->oid[colNode->oidlen - 1] > maxid))
+                         maxid = colNode->oid[colNode->oidlen - 1];
+	        fprint(f, "{ %s %d }\n\n",
+                       smiGetFirstChildNode(smiParentNode)->name, 
+                       (maxid + 1) > 128 ? (maxid + 1) : 128);
+                xfree(rowStatus);
+            } else
+ 	        fprint(f, "\n}\n\n");
 	}
     }
 }
@@ -1315,8 +1516,9 @@ static void fprintNotifications(FILE *f, SmiModule *smiModule)
 	if (smiv1) {
 	    fprint(f, "%s TRAP-TYPE\n", smiNode->name);
 	    parentNode = smiGetParentNode(smiNode);
-	    while ((!parentNode->name) ||
-		   (!parentNode->oid[parentNode->oidlen-1])) {
+	    while ((parentNode->oidlen > 0) &&
+		   ((!parentNode->name) ||
+		    (!parentNode->oid[parentNode->oidlen-1]))) {
 		parentNode = smiGetParentNode(parentNode);
 	    }
 	    fprintSegment(f, INDENT, "ENTERPRISE", INDENTVALUE, 0);
@@ -1555,10 +1757,10 @@ static void fprintModuleCompliances(FILE *f, SmiModule *smiModule)
 				      INDENTVALUE, smiv1);
 			fprint(f, "\n");
 			if (smiOption->description) {
-			    fprintMultilineString(f, smiOption->description,
+			    fprintMultilineString2(f, smiOption->description,
 						  smiv1);
 			} else {
-			    fprintMultilineString(f, "...", smiv1);
+			    fprintMultilineString2(f, "...", smiv1);
 			}
 			fprint(f, "\n");
 		    }
@@ -1610,11 +1812,11 @@ static void fprintModuleCompliances(FILE *f, SmiModule *smiModule)
 				      INDENTVALUE, smiv1);
 			fprint(f, "\n");
 			if (smiRefinement->description) {
-			    fprintMultilineString(f,
+			    fprintMultilineString2(f,
 						  smiRefinement->description,
 						  smiv1);
 			} else {
-			    fprintMultilineString(f, "...", smiv1);
+			    fprintMultilineString2(f, "...", smiv1);
 			}
 			fprint(f, "\n");
 		    }
@@ -1660,13 +1862,26 @@ static void fprintModuleCompliances(FILE *f, SmiModule *smiModule)
 
 static void dumpSmi(FILE *f, SmiModule *smiModule)
 {
+    if (smiModule->language == SMI_LANGUAGE_SPPI) /* PIB to MIB conversion */
+        pibtomib = 1;
+    else
+        pibtomib = 0;
+
     createImportList(smiModule);
     
     fprint(f, "--\n");
     fprint(f, "-- This %s module has been generated by smidump "
 	   SMI_VERSION_STRING ". Do not edit.\n", smiv1 ? "SMIv1" : "SMIv2");
     fprint(f, "--\n\n");
-    fprint(f, "%s DEFINITIONS ::= BEGIN\n\n", smiModule->name);
+    if (pibtomib && smiv1) {
+        fprint(f, "--\n");
+        fprint(f, "-- WARNING: SPPI to SMIv1 conversion is not explicitly " \
+          "supported.\n");
+        fprint(f, "-- Expect flawed output.\n");
+        fprint(f, "--\n\n");
+    }
+    fprint(f, "%s%s DEFINITIONS ::= BEGIN\n\n", smiModule->name,
+           (pibtomib ? "-MIB" : ""));
 	
     fprintImports(f);
     fprintModuleIdentity(f, smiModule);
@@ -1677,7 +1892,8 @@ static void dumpSmi(FILE *f, SmiModule *smiModule)
     fprintGroups(f, smiModule);
     fprintModuleCompliances(f, smiModule);
     
-    fprint(f, "END -- end of module %s.\n", smiModule->name);
+    fprint(f, "END -- end of module %s%s.\n", smiModule->name,
+           (pibtomib ? "-MIB" : ""));
 
     freeImportList();
 }
@@ -1702,6 +1918,10 @@ static void dumpSmiV1(int modc, SmiModule **modv, int flags, char *output)
 
     for (i = 0; i < modc; i++) {
 	dumpSmi(f, modv[i]);
+	if (fflush(f) || ferror(f)) {
+	    perror("smidump: write error");
+	    exit(1);
+	}
     }
 
     if (output) {
@@ -1730,6 +1950,11 @@ static void dumpSmiV2(int modc, SmiModule **modv, int flags, char *output)
     for (i = 0; i < modc; i++) {
 	dumpSmi(f, modv[i]);
     }
+    
+    if (fflush(f) || ferror(f)) {
+	perror("smidump: write error");
+	exit(1);
+    }
 
     if (output) {
 	fclose(f);
