@@ -5,11 +5,12 @@
  *
  * Copyright (c) 1999 Frank Strauss, Technical University of Braunschweig.
  * Copyright (c) 1999 J. Schoenwaelder, Technical University of Braunschweig.
+ * Copyright (c) 2002 J. Schoenwaelder, University of Osnabrueck.
  *
  * See the file "COPYING" for information on usage and redistribution
  * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
  *
- * @(#) $Id: dump-tree.c,v 1.28 2002/03/05 11:57:28 schoenw Exp $
+ * @(#) $Id: dump-tree.c,v 1.37 2003/07/14 22:58:42 schoenw Exp $
  */
 
 #include <config.h>
@@ -25,9 +26,10 @@ static int pmodc = 0;
 static SmiModule **pmodv = NULL;
 
 static int ignoreconformance = 0;
+static int ignoreleafs = 0;
+static int full = 0;
 static int compact = 0;
 
-
 static char *getFlags(SmiNode *smiNode)
 {
 
@@ -44,6 +46,12 @@ static char *getFlags(SmiNode *smiNode)
 	return "rwn";
     case SMI_ACCESS_NOT_IMPLEMENTED:
 	return "---";
+    case SMI_ACCESS_INSTALL:
+	return "-i-";
+    case SMI_ACCESS_INSTALL_NOTIFY:
+	return "-in";
+    case SMI_ACCESS_REPORT_ONLY:
+	return "--r";
     }
 
     return "";
@@ -144,38 +152,113 @@ static void fprintObjects(FILE *f, SmiNode *smiNode)
 
 
 
-static int pruneSubTree(SmiNode *smiNode)
+static int isPartOfLoadedModules(SmiNode *smiNode)
 {
     SmiModule *smiModule;
-    SmiNode   *childNode;
-    int       i;
+    int i;
+    
+    smiModule = smiGetNodeModule(smiNode);
 
-    const int mask = (SMI_NODEKIND_GROUP | SMI_NODEKIND_COMPLIANCE);
+    for (i = 0; i < pmodc; i++) {
+	if (strcmp(pmodv[i]->name, smiModule->name) == 0) {
+	    return 1;
+	}
+    }
+    return 0;
+}
+
+/*
+ * The following function pruneSubTree() is tricky. There are some
+ * interactions between the supported options. See the detailed
+ * comments below. Good examples to test the implemented behaviour
+ * are:
+ *
+ * smidump -u -f tree --tree-no-leaf IF-MIB ETHER-CHIPSET-MIB
+ *
+ * (And the example above does _not_ work in combination with
+ * --tree-no-conformance so the code below is still broken.)
+ */
+
+static int pruneSubTree(SmiNode *smiNode)
+{
+    SmiNode   *childNode;
+
+    const int confmask = (SMI_NODEKIND_GROUP | SMI_NODEKIND_COMPLIANCE);
+    const int leafmask = (SMI_NODEKIND_GROUP | SMI_NODEKIND_COMPLIANCE
+			  | SMI_NODEKIND_COLUMN | SMI_NODEKIND_SCALAR
+			  | SMI_NODEKIND_ROW | SMI_NODEKIND_NOTIFICATION);
 
     if (! smiNode) {
 	return 1;
     }
 
-    if (ignoreconformance && (smiNode->nodekind & mask)) {
+    /*
+     * First, prune all nodes which the user has told us to ignore.
+     * In the case of ignoreleafs, we have to special case nodes with
+     * an unknown status (which actually represent OBJECT-IDENTITY
+     * definitions). More special case code is needed to exclude
+     * module identity nodes.
+     */
+
+    if (ignoreconformance && (smiNode->nodekind & confmask)) {
 	return 1;
     }
 
-    smiModule = smiGetNodeModule(smiNode);
-    for (i = 0; i < pmodc; i++) {
-	if (strcmp(pmodv[i]->name, smiModule->name) == 0) {
-	    if (!ignoreconformance || !smiGetFirstChildNode(smiNode)) {
-		return 0;
+    if (ignoreleafs) {
+	if (smiNode->nodekind & leafmask) {
+	    return 1;
+	}
+	if (smiNode->nodekind == SMI_NODEKIND_NODE
+	    && smiNode->status != SMI_STATUS_UNKNOWN) {
+	    SmiModule *smiModule = smiGetNodeModule(smiNode);
+	    if (smiModule && smiNode != smiGetModuleIdentityNode(smiModule)) {
+		return 1;
 	    }
 	}
     }
 
+    /*
+     * Next, generally do not prune nodes that belong to the set of
+     * modules we are looking at.
+     */
+
+    if (isPartOfLoadedModules(smiNode)) {
+	if (!ignoreconformance || !smiGetFirstChildNode(smiNode)) {
+	    return 0;
+	}
+    }
+
+    /*
+     * Finally, prune all nodes where all child nodes are pruned.
+     */
+
     for (childNode = smiGetFirstChildNode(smiNode);
 	 childNode;
 	 childNode = smiGetNextChildNode(childNode)) {
+
+	/*
+	 * In the case of ignoreleafs, we have to peek at the child
+	 * nodes. Otherwise, we would prune too much. we still want to
+	 * see the path to the leafs we have pruned away. This also
+	 * interact with the semantics of ignoreconformance since we
+	 * still want in combination with ignoreleafs to see the path
+	 * to the pruned conformance leafs.
+	 */
+	
+	if (ignoreleafs && (childNode->nodekind & leafmask)) {
+	    if (isPartOfLoadedModules(childNode)) {
+		if (ignoreconformance && (childNode->nodekind & confmask)) {
+		    return 1;
+		}
+		return 0;
+	    }
+	}
+ 
 	if (! pruneSubTree(childNode)) {
 	    return 0;
 	}
     }
+
     return 1;
 }
 
@@ -197,8 +280,10 @@ static void fprintSubTree(FILE *f, SmiNode *smiNode,
 	switch (smiNode->nodekind) {
 	case SMI_NODEKIND_SCALAR:
 	case SMI_NODEKIND_COLUMN:
-	    c = prefix[prefixlen-1];
-	    prefix[prefixlen-1] = getStatusChar(smiNode->status);
+	    if (prefixlen > 0) {
+	        c = prefix[prefixlen-1];
+		prefix[prefixlen-1] = getStatusChar(smiNode->status);
+	    }
 	    type_name = getTypeName(smiNode);
 	    if (type_name) {
 		fprintf(f, "%s-- %s %-*s %s(%u)\n",
@@ -210,12 +295,12 @@ static void fprintSubTree(FILE *f, SmiNode *smiNode,
 			smiNode->oid[smiNode->oidlen-1]);
 		xfree(type_name);
 	    }
-	    if (c) {
+	    if (prefixlen > 0 && c) {
 		prefix[prefixlen-1] = c;
 	    }
 	    break;
 	case SMI_NODEKIND_ROW:
-	    if (prefixlen) {
+	    if (prefixlen > 0) {
 		c = prefix[prefixlen-1];
 		prefix[prefixlen-1] = getStatusChar(smiNode->status);
 	    }
@@ -240,12 +325,12 @@ static void fprintSubTree(FILE *f, SmiNode *smiNode,
 		break;	    
 	    }
 	    fprintf(f, "]\n");
-	    if (c) {
+	    if (prefixlen > 0 && c) {
 		prefix[prefixlen-1] = c;
 	    }
 	    break;
 	case SMI_NODEKIND_NOTIFICATION:
-	    if (prefixlen) {
+	    if (prefixlen > 0) {
 		c = prefix[prefixlen-1];
 		prefix[prefixlen-1] = getStatusChar(smiNode->status);
 	    }
@@ -254,22 +339,22 @@ static void fprintSubTree(FILE *f, SmiNode *smiNode,
 		    smiNode->oid[smiNode->oidlen-1]);
 	    fprintObjects(f, smiNode);
 	    fprintf(f, "]\n");
-	    if (c) {
+	    if (prefixlen > 0 && c) {
 		prefix[prefixlen-1] = c;
 	    }
 	    break;
 	default:
-	    if (prefixlen) {
+	    if (prefixlen > 0) {
 		c = prefix[prefixlen-1];
 		prefix[prefixlen-1] = getStatusChar(smiNode->status);
 	    }
 	    if (smiNode->oid)
-		if (prefixlen) {
+		if (prefixlen > 0) {
 		    fprintf(f, "%s--%s(%u)\n", prefix,
 			    smiNode->name ? smiNode->name : " ",
 			    smiNode->oid[smiNode->oidlen-1]);
 		} else {
-		    int j;
+		    unsigned int j;
 		    fprintf(f, "%s--%s(", prefix,
 			    smiNode->name ? smiNode->name : " ");
 		    for (j = 0; j < smiNode->oidlen; j++) {
@@ -278,9 +363,9 @@ static void fprintSubTree(FILE *f, SmiNode *smiNode,
 		    fprintf(f, ")\n");
 		}
 	    else
-		fprintf(f, "%s--%s(??)\n", prefix,
+		fprintf(f, "%s--%s(?)\n", prefix,
 			smiNode->name ? smiNode->name : " ");
-	    if (c) {
+	    if (prefixlen > 0 && c) {
 		prefix[prefixlen-1] = c;
 	    }
 	}
@@ -309,9 +394,10 @@ static void fprintSubTree(FILE *f, SmiNode *smiNode,
 		continue;
 	    }
 	    i++;
-	    if ((childNode->nodekind != SMI_NODEKIND_COLUMN
-		&& childNode->nodekind != SMI_NODEKIND_SCALAR)
-		|| (lastNodeKind != childNode->nodekind)) {
+	    if (! compact &&
+		((childNode->nodekind != SMI_NODEKIND_COLUMN
+		  && childNode->nodekind != SMI_NODEKIND_SCALAR)
+		 || (lastNodeKind != childNode->nodekind))) {
 		fprintf(f, "%s  |\n", prefix);
 	    }
 	    newprefix = xmalloc(strlen(prefix)+10);
@@ -339,7 +425,7 @@ static void fprintTree(FILE *f)
     
     smiNode = smiGetNode(NULL, "iso");
 
-    if (compact) {
+    if (! full) {
 	do {
 	    for (childNode = smiGetFirstChildNode(smiNode), cnt = 0, nextNode = NULL;
 		 childNode;
@@ -404,6 +490,11 @@ static void dumpTree(int modc, SmiModule **modv, int flags, char *output)
 	}
     }
 
+    if (fflush(f) || ferror(f)) {
+	perror("smidump: write error");
+	exit(1);
+    }
+
     if (output) {
 	fclose(f);
     }
@@ -414,10 +505,14 @@ static void dumpTree(int modc, SmiModule **modv, int flags, char *output)
 void initTree()
 {
     static SmidumpDriverOption opt[] = {
-	{ "ignore-conformance", OPT_FLAG, &ignoreconformance, 0,
-	  "ignore conformance definitions"},
-	{ "no-full-path", OPT_FLAG, &compact, 0,
-	  "do not show the full path to the root"},
+	{ "no-conformance", OPT_FLAG, &ignoreconformance, 0,
+	  "do not show conformance nodes"},
+	{ "no-leafs", OPT_FLAG, &ignoreleafs, 0,
+	  "do not show leaf nodes"},
+	{ "full-root", OPT_FLAG, &full, 0,
+	  "generate the full path to the root"},
+	{ "compact", OPT_FLAG, &compact, 0,
+	  "generate a more compact representation"},
         { 0, OPT_END, 0, 0 }
     };
     
